"""
A collection of Widgets and Listeners designed specifically for the ECS system.
"""

from bear_hug.bear_utilities import BearECSException
from bear_hug.ecs import Entity
from bear_hug.event import BearEvent
from bear_hug.widgets import Layout


class ECSLayout(Layout):
    """
    A Layout of entities.
    
    It is controlled entirely by events. Although Layour methods like
    `add_child` and `move_child` are not overloaded, their use is discouraged.
    
    In all events, entity IDs, not the actual Entity objects, are passed. Event
    conventions are as following:
    
    `BearEvent(event_type='ecs_move', event_value=(entity, x, y))`.
    Announces that the widget of the entity in question should be moved
    to (x; y). Behaviour for widgets trying to leave map edges is left to
    be determined by ECSLayout or its children.
    
    `BearEvent(event_type='ecs_remove', event_value=entity)`.
    Announces that the widget of the entity in question should be removed from
    the ECSLayout. Does not cause or imply the destruction of entity object.
    
    `BearEvent(event_type='ecs_add', event_value=(entity, x, y))`.
    Announces that the widget of the entity in question should be added to the
    ECSLayout at (x;y). The emission of this event implies the existence of the
    entity and its widget.
    
    This widget also provides the collision detection for all widgets within it.
    If a widget attempts to move into the position occupied by some other
    widget, the movement is not blocked, but the 'ecs_collision' event(s) get
    emitted. The event follows the following convention:
    
    `BearEvent(event_type='ecs_collision', event_value=(widget_moved.id,
                                                    widget_collided_into.id))`
    If the widget enters the screen border (ie will go outside it the next time
    it moves, assuming it keeps the direction), the event takes the following
    form:
    `BearEvent(event_type='ecs_collision', event_value=(widget_moved.id,
                                                        None))`
    """
    
    def __init__(self, chars, colors):
        super().__init__(chars, colors)
        self.entities = {}
        self.widgets = {}
        self.need_redraw = False
    
    def add_entity(self, entity):
        """
        Register the entity to be displayed. Assumes that the entity has a
        widget already.
        
        The entity is not actually shown until the `'ecs_add'` event is emitted
        :return:
        """
        if not isinstance(entity, Entity):
            raise BearECSException('Cannot add non-Entity to ECSLayout')
        self.entities[entity.id] = entity
        self.widgets[entity.id] = entity.widget.widget
        
    def on_event(self, event):
        # React to the events
        r = []
        if event.event_type == 'ecs_move':
            entity_id, x, y = event.event_value
            self.move_child(self.widgets[entity_id], (x, y))
            self.need_redraw = True
            # Checking if collision events need to be emitted
            # Check for collisions with border
            if x == 0 or x+self.entities[entity_id].widget.size[0]\
                 == len(self.chars[0]) or y == 0 or \
                 y + self.entities[entity_id].widget.size[1] == len(self.chars):
                r.append(BearEvent(event_type='ecs_collision',
                                   event_value=(entity_id, None)))
            else:
                collided = set()
                for y_offset in range(self.entities[entity_id].widget.size[1]):
                    for x_offset in range(self.entities[entity_id].widget.size[0]):
                        for other_widget in self._child_pointers[y+y_offset]\
                            [x+x_offset]:
                            # Child_pointers is ECS-agnostic and stores pointers
                            # to the actual widgets
                                collided.add(other_widget)
                collided_ent_ids = set()
                for child in self.entities:
                    if child != entity_id and \
                            self.entities[child].widget.widget in collided:
                        collided_ent_ids.add(child)
                for child in collided_ent_ids:
                    r.append(BearEvent('ecs_collision', (entity_id, child)))
        elif event.event_type == 'ecs_create':
            self.add_entity(event.event_value)
            self.need_redraw = True
        elif event.event_type == 'ecs_remove':
            self.remove_child(self.entities[event.event_value].widget)
            self.need_redraw = True
        elif event.event_type == 'ecs_add':
            entity_id, x, y = event.event_value
            self.add_child(self.widgets[entity_id], (x, y))
            self.need_redraw = True
        elif event.event_type == 'ecs_update':
            # Some widget has decided it's time to redraw itself
            self.need_redraw = True
        elif event.event_type == 'service' and event.event_value == 'tick_over'\
                and self.need_redraw:
            self._rebuild_self()
            self.terminal.update_widget(self)
            self.need_redraw = False
        if r:
            return r



"""
A sound system.

Currently it exports a single class called `SoundListener`. It's a Listener
wrapper around simpleaudio and wave libraries. While later the backend is likely
to change (at least to support sound formats other than `.wav`), event API is
probably gonna remain backward-compatible.
"""

from bear_hug.bear_utilities import BearSoundException
from bear_hug.widgets import Listener

import simpleaudio as sa
import wave


class SoundListener(Listener):
    """
    It doesn't listen to sounds. It listens to the *events* and plays sounds.
    
    This class is expected to be used as a singleton, *ie* there is no reason to
    have two SoundListeners active at the same time and therefore no API for it.
    
    Accepts events like this:
    
    `BearEvent(event_type='play_sound', event_value=sound_name)`
    
    If sound_name is a known sound ID, this sound is (asynchronously) played.
    Otherwise, BearSoundException is raised. Sounds can be either supplied in a
    single arg during creation, or added later via register_sound. In either
    case, for a sound either a `simpleaudio.WaveObject` or a string is expected.
    In the latter case, a string is treated as a path to a `.wav` file.
    """
    def __init__(self, sounds):
        if not isinstance(sounds, dict):
            raise BearSoundException(
                'Only a dict accepted at SoundListener creation')
        if any((not isinstance(x, str) for x in sounds)):
            raise BearSoundException('Only strings accepted as sound IDs')
        for sound_name in sounds:
            if isinstance(sounds[sound_name], sa.WaveObject):
                continue
            if isinstance(sounds[sound_name], str):
                sounds[sound_name] =  sa.WaveObject.from_wave_read(
                    wave.open(sounds[sound_name], 'rb'))
            else:
                raise BearSoundException(
                    'Sound should be either WaveObject or string')
        self.sounds = sounds
        
    def register_sound(self, sound, sound_name):
        """
        Register a new sound for this listener
        :param sound: WaveObject or str. A sound to be registered
        :param sound_name: name of this sound.
        :return:
        """
        if sound_name in self.sounds:
            raise BearSoundException(f'Duplicate sound name "{sound_name}"')
        if isinstance(sound, sa.WaveObject):
            self.sounds[sound_name] = sound
        elif isinstance(sound, str):
            self.sounds[sound_name] = sa.WaveObject.from_wave_read(
                    wave.open('dsstnmov.wav', 'rb'))
        else:
            raise BearSoundException(
                'Sound should be either WaveObject or string')
    
    def play_sound(self, sound_name):
        """
        Play a sound.
        
        In case you need to play the sound without requesting it through the
        event.
        :param sound_name: A sound to play.
        :return:
        """
        if sound_name not in self.sounds:
            raise BearSoundException(
                f'Nonexistent sound {sound_name} requested')
        self.sounds[sound_name].play()
    
    def on_event(self, event):
        if event.event_type == 'play_sound':
            self.play_sound(event.event_value)



"""
Loaders for the various ASCII-art formats
"""

from bear_hug.bear_utilities import BearException, copy_shape, rotate_list
from copy import deepcopy
import base64
import gzip
import json
import os


class ASCIILoader:
    """
    A base class for all the resource loaders. It knows how to return its chars
    and colors (or a fragment thereof), but expects the children to do their
    loading by themselves.
    """
    def __init__(self):
        self.chars = None
        self.colors = None
    
    def get_image(self):
        """
        Return the entire chars and colors of this loader
        :return:x
        """
        if not self.chars or not self.colors:
            raise BearException('Loader is empty')
        return self.chars, self.colors
    
    def get_image_region(self, x, y, xsize, ysize):
        """
        Return some region of this loader's chars and colors
        As everywhere, the coordinates start from 0.
        :return:
        """
        if not self.chars or not self.colors:
            raise BearException('Loader is empty')
        # Checking for the correct pos and size
        if x < 0 or x > len(self.chars[0]) or y < 0 or y > len(self.chars):
            raise BearException('Region outside loader boundaries')
        if x + xsize > len(self.chars[0]) or y + ysize > len(self.chars):
            raise BearException('Region too huge')
        ch = []
        co = []
        for y_offset in range(ysize):
            r = []
            c = []
            for x_offset in range(xsize):
                try:
                    r.append(self.chars[y + y_offset][x + x_offset])
                    c.append(self.colors[y + y_offset][x + x_offset])
                except IndexError:
                    print(x_offset, y_offset)
            ch.append(r)
            co.append(c)
        return ch, co


class TxtLoader(ASCIILoader):
    """
    A loader that reads a plaintext file.
    On the creation, the loader accepts a filename (anything acceptable by
    `open()`), default color and a bool `load_file`.
    The first argument controls what color would the characters be assigned. As
    `.txt` format does not store colors, all chars will be the same color. The
    latter controls whether the file is read immediately. If False (default), on
    loader creation it only checks that the file exists; if True, its contents
    are immediately loaded to `loader.chars`.
    """
    def __init__(self, filename, default_color='white', load_file=False):
        super().__init__()
        self.default_color = default_color
        if not os.path.exists(filename):
            raise ValueError('Nonexistent path {}'.format(filename))
        self.filename = filename
        if load_file:
            self._load_file()
        
    def _load_file(self):
        for line in open(self.filename):
            f = list(line.rstrip('\n'))
            if self.chars and len(f) != len(self.chars[0]):
                raise BearException('All lines should be equal length')
            if not self.chars:
                self.chars = [f]
            else:
                self.chars.append(f)
        self.colors = copy_shape(self.chars, self.default_color)
        
    def get_image(self):
        if not self.chars or not self.colors:
            self._load_file()
        return super().get_image()
        
    def get_image_region(self, x, y, xsize, ysize):
        if not self.chars or not self.colors:
            self._load_file()
        return super().get_image_region(x, y, xsize, ysize)


class XpLoader(ASCIILoader):
    """
    A loader that reads REXPaint *.xp files.
    As with the other loaders, the file is not parsed until one of the `get_*`
    methods gets called. Its existence, though, is checked on Loader creation.
    
    Most of the code is taken from MIT-licensed XPLoaderPy3, copyright
    Sean Hagar, Erwan Castioni and Gawein Le Goff.
    
    As the bear_hug widget API does not allow multi-layered widgets, `get_image`
    and `get_image_region` return the image with the only the character and
    color from the highest layer which is non-empty for a given cell. For
    getting data from layers separately, use `get_layer` and `get_layer_region`.
    
    Background colors are ignored altogether.
    """

    version_bytes = 4
    layer_count_bytes = 4

    layer_width_bytes = 4
    layer_height_bytes = 4
    layer_keycode_bytes = 4
    layer_fore_rgb_bytes = 3
    layer_back_rgb_bytes = 3
    layer_cell_bytes = layer_keycode_bytes + layer_fore_rgb_bytes + \
                       layer_back_rgb_bytes
    transparent_cell_back_r = 255
    transparent_cell_back_g = 0
    transparent_cell_back_b = 255

    # These chars are not correctly decoded by python's bytes-to-str conversion
    # and have to be processed manually in `_parse_individual_cell()`
    fix_chars = {0x01: "\u263A", 0x02: "\u263B", 0x03: "\u2665", 0x04: "\u2666",
        0x05: "\u2663", 0x06: "\u2660", 0x07: "\u2022", 0x08: "\u25D8",
        0x09: "\u25CB", 0x0a: "\u25D9", 0x0b: "\u2642", 0x0c: "\u2640",
        0x0d: "\u266A", 0x0e: "\u266B", 0x0f: "\u263C", 0x10: "\u25BA",
        0x11: "\u25C4", 0x12: "\u2195", 0x13: "\u203C", 0x14: "\u00B6",
        0x15: "\u00A7", 0x16: "\u25AC", 0x17: "\u21A8", 0x18: "\u2191",
        0x19: "\u2193", 0x1a: "\u2192", 0x1b: "\u2190", 0x1c: "\u221F",
        0x1d: "\u2194", 0x1e: "\u25B2", 0x1f: "\u25BC", 0x7f: "\u2302"}
    
    def __init__(self, filename, default_color='white'):
        super().__init__()
        if not os.path.exists(filename):
            raise ValueError('Nonexistent path {}'.format(filename))
        self.filename = filename
        self.default_color = default_color
        self.layers = []
        self.rexpaint_version = None
        self.width = None
        self.height = None
        self.layer_count = None
    
    def get_image(self):
        """
        Return chars and colors for the entire image. For each cell only the
        values from the topmost layer are used. Background colors are
        ignored altogether.
        :return:
        """
        if not self.chars:
            self._process_xp_file()
            self._get_topmost_layer()
        return super().get_image()

    def get_image_region(self, x, y, xsize, ysize):
        """
        Return chars and colors for the image region. For each cell only the
        values from the topmost layer are used. Background colors are
        ignored altogether.
        :param x:
        :param y:
        :param xsize:
        :param ysize:
        :return:
        """
        if not self.chars:
            self._process_xp_file()
            self._get_topmost_layer()
        return super().get_image_region(x, y, xsize, ysize)
    
    def get_layer(self, layer):
        """
        Get chars and (foreground) colors for the entire image layer. This
        method does not check layer size and returns whatever available.
        
        By default, the REXPaint creates layers the size of the entire image, so
        this shouldn't be much of an issue.
        :param layer:
        :return:
        """
        if not self.layers:
            self._process_xp_file()
        if layer >= self.layer_count:
            raise BearException('Nonexistent layer in XpLoader')
        return deepcopy(self.layers[layer][0]), deepcopy(self.layers[layer][1])
    
    def get_layer_region(self, layer, x, y, xsize, ysize):
        if not self.layers:
            self._process_xp_file()
        if layer >= self.layer_count:
            raise BearException('Nonexistent layer in XpLoader')
        # Shamelessly copypasted from ASCIILoader.get_image_region
        ch = []
        co = []
        for y_offset in range(ysize):
            r = []
            c = []
            for x_offset in range(xsize):
                r.append(self.layers[layer][0][y + y_offset][x + x_offset])
                c.append(self.layers[layer][1][y + y_offset][x + x_offset])
            ch.append(r)
            co.append(c)
        return ch, co

    def _process_xp_file(self):
        gz_handle = gzip.open(self.filename)
        line = gz_handle.read()
        self._load_xp_string(line)
        
    def _get_topmost_layer(self):
        if self.layer_count == 1:
            self.chars = deepcopy(self.layers[0][0])
            self.colors = deepcopy(self.layers[0][1])
        else:
            self.chars = [ [' ' for x in range(self.width)]
                           for y in range(self.height)]
            self.colors = copy_shape(self.chars, None)
            for row in range(self.height):
                for column in range(self.width):
                    for layer in self.layers[::-1]:
                        if layer[0][row][column] != ' ':
                            self.chars[row][column] = layer[0][row][column]
                            self.colors[row][column] = layer[1][row][column]
                            break
        
    # All code from here to the end of the class is adapted from XPLoaderPy3
    # Looks like a mess of crutches on top of other crutches, but I'm not really
    # up to building an elegant solution right now.
    def _load_xp_string(self, file_string, reverse_endian=True):
        """
        Parse REXpaint string and populate self.layers
        :param file_string:
        :param reverse_endian:
        :return:
        """
        offset = 0
        version = file_string[offset: offset + self.version_bytes]
        offset += self.version_bytes
        layer_count = file_string[offset: offset + self.layer_count_bytes]
        offset += self.layer_count_bytes
        if reverse_endian:
            version = version[::-1]
            layer_count = layer_count[::-1]
        self.version = int(base64.b16encode(version), 16)
        self.layer_count = int(base64.b16encode(layer_count), 16)
        current_largest_width = 0
        current_largest_height = 0
        for layer in range(self.layer_count):
            # slight lookahead to figure out how much data to feed load_layer
            this_layer_width = file_string[offset:offset +
                                           self.layer_width_bytes]
            this_layer_height = file_string[
                            offset + self.layer_width_bytes:offset +
                            self.layer_width_bytes + self.layer_height_bytes]
            
            if reverse_endian:
                this_layer_width = this_layer_width[::-1]
                this_layer_height = this_layer_height[::-1]
                
            this_layer_width = int(base64.b16encode(this_layer_width), 16)
            this_layer_height = int(base64.b16encode(this_layer_height), 16)
            current_largest_width = max(current_largest_width, this_layer_width)
            current_largest_height = max(current_largest_height,
                                         this_layer_height)
        
            layer_data_size = self.layer_width_bytes + self.layer_height_bytes\
                + (self.layer_cell_bytes * this_layer_width * this_layer_height)
            layer_data = self._parse_layer(
                file_string[offset:offset + layer_data_size], reverse_endian)
            self.layers.append(layer_data)
            offset += layer_data_size

    def _parse_layer(self, layer_string, reverse_endian=True):
        """
        Parse a file portion for a single layer
        :param layer_string:
        :param reverse_endian:
        :return:
        """
        offset = 0
        width = layer_string[offset:offset + self.layer_width_bytes]
        offset += self.layer_width_bytes
        height = layer_string[offset:offset + self.layer_height_bytes]
        offset += self.layer_height_bytes
    
        if reverse_endian:
            width = width[::-1]
            height = height[::-1]
    
        self.width = int(base64.b16encode(width), 16)
        self.height = int(base64.b16encode(height), 16)
        cells = []
        for x in range(self.width):
            row = []
            for y in range(self.height):
                cell_data_raw = layer_string[offset:offset +
                                                    self.layer_cell_bytes]
                cell_data = self._parse_individual_cell(cell_data_raw,
                                                        reverse_endian)
                row.append(cell_data)
                offset += self.layer_cell_bytes
            cells.append(row)
        cells = rotate_list(cells)
        chars = copy_shape(cells, None)
        colors = copy_shape(cells, self.default_color)
        for r in range(len(cells)):
            for c in range(len(cells[0])):
                chars[r][c] = cells[r][c][0]
                colors[r][c] = cells[r][c][1]
        return chars, colors

    def _parse_individual_cell(self, cell_string, reverse_endian=True):
        """
        Process individual cell data
        :param cell_string:
        :param reverse_endian:
        :return:
        """
        offset = 0
        keycode = cell_string[offset:offset + self.layer_keycode_bytes]
        if reverse_endian:
            keycode = keycode[::-1]
        keycode = int(base64.b16encode(keycode), 16)
        # Processing characters that are redefined by IBM437 from ASCII control
        # sequences and are thus not correctly decoded by `.decode('cp437')`
        if keycode in self.fix_chars:
            char = self.fix_chars[keycode]
        else:
            char = bytes([keycode]).decode('cp437')
        if not char or char == '\x00':
            char = ' '
        offset += self.layer_keycode_bytes
        fore_r = int(base64.b16encode(cell_string[offset:offset + 1]), 16)
        offset += 1
        fore_g = int(base64.b16encode(cell_string[offset:offset + 1]), 16)
        offset += 1
        fore_b = int(base64.b16encode(cell_string[offset:offset + 1]), 16)
        offset += 1
        # Covering cases when we have like '0x4' and '0xAB' in color
        # panning shorter values with zeroes to max size
        rgb = [str(hex(x)).split('x')[1] for x in (fore_r, fore_g, fore_b)]
        length = max((len(x) for x in rgb))
        if length >= 2:
            for index in range(len(rgb)):
                if len(rgb[index]) < length:
                    rgb[index] = '0'*(length - len(rgb[index])) + rgb[index]
        color = '#' + ''.join(rgb)
        # `Back_*` values are ignored, but this code isn't removed in case they
        # are needed later.
        back_r = int(base64.b16encode(cell_string[offset:offset + 1]), 16)
        offset += 1
        back_g = int(base64.b16encode(cell_string[offset:offset + 1]), 16)
        offset += 1
        back_b = int(base64.b16encode(cell_string[offset:offset + 1]), 16)
        offset += 1
        return char, color


class Atlas:
    """
    An image atlas.
    
    An instance of this class accepts a Loader instance and a path to the JSON
    file. The latter is parsed immediately.
    """
    def __init__(self, loader, json_file):
        self.loader = loader
        # A dict of {name: (x, y, xsize, ysize)}
        self.elements = {}
        for item in json.load(open(json_file)):
            self.elements[item['name']] = (item['x'], item['y'],
                                        item['xsize'], item['ysize'])
        
    def get_element(self, name):
        """
        Return an element with a given name
        :param name:
        :return:
        """
        return self.loader.get_image_region(*self.elements[name])



#! /usr/bin/env python3.6

from bear_hug import BearTerminal, BearLoop
from bear_utilities import copy_shape
from event import BearEventDispatcher
from widgets import InputScrollable, ClosingListener, Layout,\
    Label, Widget, FPSCounter
from resources import XpLoader, Atlas

import os

class ElementBox(Layout):
    """
    A box for a given widget.
    
    Consists of widget itself, two lines of empty space around it and a '#'-box
    around *that*. The upper border of the box also includes a title.
    """
    def __init__(self, widget, name='Widget', color='#ff999999'):
        if widget.width + 4 >= len(name) + 1:
            box = self.generate_box(widget.width+2, widget.height+2,
                                                color)
        else:
            box = self.generate_box(len(name), widget.height+2, color)
        super().__init__(*box)
        self.add_child(widget, pos=(2, 2))
        self.add_child(Label(name, color='green'), pos=(1, 0))
        self._rebuild_self()
    
    @staticmethod
    def generate_box(width, height, color):
        """
        Return a #-bound box of a given (internal) size
        :param width:
        :param height:
        :return:
        """
        chars = []
        chars.append(['#' for x in range(width+2)])
        for y in range(height):
            chars.append(['#'] + [' ' for x in range(width)] + ['#'])
        chars.append(['#' for x in range(width+2)])
        colors = copy_shape(chars, color)
        return chars, colors

    
t = BearTerminal(size='46x52', title='Atlas',
                 filter=['keyboard', 'mouse'])
dispatcher = BearEventDispatcher()
loop = BearLoop(t, dispatcher)
dispatcher.register_listener(ClosingListener(), ['misc_input', 'tick'])
t.start()


atlas = Atlas(XpLoader(os.path.dirname(__file__)+'/demo_assets/test_atlas.xp'),
              os.path.dirname(__file__)+'/demo_assets/test_atlas.json')
elements = []
positions = []
names = []
x = 1
y = 1
y_step = 0
for element in sorted(atlas.elements.keys()):
    w = ElementBox(Widget(*atlas.get_element(element)), name=element)
    elements.append(w)
    if x + w.width > 45:
        y += y_step
        x = 1
        y_step = 0
    positions.append((x, y))
    x += w.width + 1
    if w.height + 1 >= y_step:
        y_step = w.height + 1
view_height = y+y_step if y+y_step > 50 else 50
chars = [[' ' for _ in range(45)] for _ in range(view_height)]
colors = copy_shape(chars, 'white')
element_view = InputScrollable(chars, colors, view_pos=(0, 0),
                               view_size=(45, 50), right_bar=True)
for index, widget in enumerate(elements):
    element_view.add_child(widget, positions[index])
dispatcher.register_listener(element_view, ['tick', 'key_down', 'service'])
dispatcher.register_listener(element_view.scrollable,
                             ['tick', 'service'])
t.add_widget(element_view, pos=(0, 0))
t.add_widget(FPSCounter(), pos=(0, 51))
loop.run()



"""
Entity-component system.

Entities are just the sets of components, nothing more. The major way for them
to be used should be components calling something like
`self.owner.that_other_component.do_stuff` or emitting `ecs_*` events.

The creation of a new Entity is announced by the following event:
`BearEvent(event_type='ecs_create', event_value=entity)`

It is the only event type that uses the actual entity object, not its ID, as the
event_value. When this event is emitted, the entity should be ready to work,
in particular, all its components should be subscribed to the appropriate events
"""

from bear_hug.bear_utilities import BearECSException
from bear_hug.widgets import Widget, Listener
from bear_hug.event import BearEvent


class Entity:
    """
    A root entity class.
    """
    
    def __init__(self, id='Default ID', components=[]):
        self.components = []
        self.id = id
        for component in components:
            self.add_component(component)
        
    def add_component(self, component):
        """
        Adds component to the Entity __dict__
        Raises exception if the Component.name corresponds to one of the already
        existing elements of __dict__ and isn't in self.components. The latter
        check is to allow overwriting components while preventing them from
        overwriting the builtin properties.
        :param component:
        :return:
        """
        if not isinstance(component, Component):
            raise BearECSException('Only Component instance can be added' +
                                   ' as an entity\'s component')
        if component.name not in self.components and \
                component.name in self.__dict__:
            raise BearECSException('Cannot add component' +
                   '{} that shadows builtin attribute'.format(component.name))
        self.__dict__[component.name] = component
        self.components.append(component.name)
        component.owner = self
    
    def remove_component(self, component_name):
        if component_name in self.components:
            del(self.components[component_name])
        else:
            raise BearECSException('Cannot remove component ' +
                          '{} that Entity doesn\'t have'.format(component_name))
        
    
class Component(Listener):
    """
    A root component class.
    
    Component name is expected to be the same between all components of the same
    class.
    
    Component inherits from Listener and is therefore able to receive and return
    BearEvents. Of course, it needs the correct subscriptions to actually get
    them.
    """
    def __init__(self, dispatcher, name='Root component', owner=None):
        super().__init__()
        if not name:
            raise BearECSException('Cannot create a component without a name')
        self.dispatcher = dispatcher
        self.name = name
        self.owner = None
        self.set_owner(owner)
            
    def set_owner(self, owner):
        """
        Registers a component owner.
        
        This is only useful if the component is passed from one owner to
        another, or if the component is created with the `owner` argument.
        This method calls owner's `add_component`
        :param owner:
        :return:
        """
        if owner:
            if not isinstance(owner, Entity):
                raise BearECSException('Only an Entity can be Component owner')
            owner.add_component(self)
        
    def on_event(self, event):
        """
        Component's event callback.
        :param event:
        :return:
        """
        pass


class WidgetComponent(Component):
    """
    Widget as a component.
    
    This component is an ECS wrapper around the Widget object. Since Widgets
    can accept events and it is sometimes reasonable to keep some event logic in
    the Widget instead of Components (ie to keep animation running), its
    `on_event` method simply passes the events to the Widget. It also supports
    `height`, `width` and `size` properties, also by calling widget's ones.
    """
    
    def __init__(self, dispatcher, widget, owner=None):
        if not isinstance(widget, Widget):
            raise TypeError('A widget is not actually a Widget')
        super().__init__(dispatcher=dispatcher, name='widget', owner=owner)
        self.widget = widget
        if self.dispatcher:
            self.dispatcher.register_listener(self.widget, 'tick')

    def on_event(self, event):
        return self.widget.on_event(event)

    @property
    def height(self):
        return self.widget.height

    @property
    def width(self):
        return self.widget.width

    @property
    def size(self):
        return self.widget.size


class PositionComponent(Component):
    """
    A component responsible for positioning Widget on ECSLayout.
    
    It has x and y coordinates, as well as vx and vy speed components.
    Coordinates are given in tiles and speed is in tiles per second.
    """
    def __init__(self, dispatcher, x=0, y=0, vx=0, vy=0, owner=None):
        super().__init__(dispatcher, name='position', owner=owner)
        self._x = x
        self._y = y
        self.vx = vx
        if self.vx:
            self.x_delay = abs(1/self.vx)
        self.x_waited = 0
        self.vy = vy
        if self.vy:
            self.y_delay = abs(1/self.vy)
        self.y_waited = 0
        if self.dispatcher:
            dispatcher.register_listener(self, 'tick')
        
    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y
    
    def move(self, x, y, emit_event=True):
        """
        Move the Entity to a specified position.
        :param x: x
        :param y: y
        :param emit_event: Whether to emit an 'esc_move' event. There are a few
        cases (ie setting the coordinates after the component is created, but
        before the entity is added to the terminal) where this is undesirable.
        :return:
        """
        self._x = x
        self._y = y
        if emit_event:
            self.dispatcher.add_event(BearEvent(event_type='ecs_move',
                                                event_value=(self.owner.id,
                                                             self._x,
                                                             self._y)))

    def relative_move(self, dx, dy, emit_event=True):
        """
        Move the Entity to a specified position.
        :param x: x
        :param y: y
        :param emit_event: Whether to emit an 'esc_move' event. There are a few
        cases (ie setting the coordinates after the component is created, but
        before the entity is added to the terminal) where this is undesirable.
        :return:
        """
        self.move(self.x+dx, self.y+dy, emit_event=emit_event)
        
    def on_event(self, event):
        if event.event_type == 'tick':
            # Move
            if self.vx or self.vy:
                self.x_waited += event.event_value
                self.y_waited += event.event_value
                if self.vx and self.x_waited > self.x_delay:
                    new_x = self.x + 1 if self.vx > 0 else self.x - 1
                    self.x_waited = 0
                else:
                    new_x = self.x
                if self.vy and self.y_waited > self.y_delay:
                    new_y = self.y + 1 if self.vy > 0 else self.y - 1
                    self.y_waited = 0
                else:
                    new_y = self.y
                if not self.x == new_x or not self.y == new_y:
                    self.move(new_x, new_y)
                

class SpawnerComponent(Component):
    """
    A component responsible for creating other entities. A current
    implementation can produce only a single Entity type; this will be fixed
    when entity factories are up and running.
    
    :param to_spawn: A callable that returns an Entity to be created. The entity
    needs to have `widget` and `position` components.
    :param relative_pos: a starting position of a spawned Entity, relative to
    self.
    """
    # TODO: accept entity factories
    def __init__(self, dispatcher, to_spawn, relative_pos=(0, 0), owner=None):
        super().__init__(dispatcher, name='spawner', owner=owner)
        self.to_spawn = to_spawn
        self.relative_pos = relative_pos
        self.id_count = 0
        
    def create_entity(self):
        entity = self.to_spawn()
        for component in (entity.__dict__[c] for c in entity.components):
            component.dispatcher = self.dispatcher
        entity.id += str(self.id_count)
        self.id_count += 1
        entity.position.move(self.owner.position.x + self.relative_pos[0],
                             self.owner.position.y + self.relative_pos[1],
                             emit_event=False)
        self.dispatcher.add_event(BearEvent(event_type='ecs_create',
                                            event_value=entity))
        self.dispatcher.add_event(BearEvent(event_type='ecs_add',
                                            event_value=(entity.id,
                                                         entity.position.x,
                                                         entity.position.y)))



"""
Various useful Widget and Listener classes
These widgets and listeners are usable outside the ECS and should be sufficient
for simpler games and apps. However, for the sake of clearer architecture,
entities are recommended.
"""

from bear_hug.bear_hug import BearTerminal
from bear_hug.bear_utilities import shapes_equal, blit, copy_shape, slice_nested,\
    BearException, BearLayoutException
from bear_hug.event import BearEvent
from collections import deque
from time import time


class Widget:
    """
    The base class for things that can be placed on the terminal.
    This class is inactive and is intended to be either inherited from or used
    for purely decorative non-animated objects. Event processing and animations
    are covered by its subclasses; while it has `on_event()` method, it does
    nothing. This allows Widgets to work without subscribing to the queue and
    saves some tacts on not redrawing them unless the Widget itself considers it
    necessary.

    Accepted parameters:
    `chars`: a list of unicode characters
    `colors`: a list of colors. Anything that is accepted by terminal.color()
    goes here (a color name or a 0xAARRGGBB/0xRRGGBB/0xRGB/0xARGB integer).

    `chars` and `colors` should be exactly the same shape, otherwise the
    BearException is raised.
    """
    
    def __init__(self, chars, colors):
        if not isinstance(chars, list) or not isinstance(colors, list):
            raise BearException('Chars and colors should be lists')
        if not shapes_equal(chars, colors):
            raise BearException('Chars and colors should have the same shape')
        self.chars = chars
        self.colors = colors
        # A widget may want to know about the terminal it's attached to
        self._terminal = None
        # Or a parent
        self._parent = None
        
    def on_event(self, event):
        # Root widget does not raise anything here, because Widget() can be
        # erroneously subscribed to a queue. While useless, that's not really a
        # fatal error.
        pass
    
    @property
    def terminal(self):
        return self._terminal
    
    @terminal.setter
    def terminal(self, value):
        if value and not isinstance(value, BearTerminal):
            raise BearException('Only a BearTerminal can be set as ' +
                                'Widget.terminal')
        self._terminal = value
        
    @property
    def parent(self):
        return self._parent
    
    @parent.setter
    def parent(self, value):
        if value and not isinstance(value, (Widget, BearTerminal)):
            raise BearException(
                'Only a widget or terminal can be a widget\'s parent')
        self._parent = value
        
    @property
    def height(self):
        return len(self.chars)
    
    @property
    def width(self):
        return len(self.chars[0])
    
    @property
    def size(self):
        return len(self.chars[0]), len(self.chars)
        
    def flip(self, axis):
        """
        Flip a widget along one of the axes.
        
        Rotates chars and colors either horizontally ('x' or 'horizontal') or
        vertically ('y' or 'vertical'). Note that this method has **extremely**
        limited uses: first, it only affects chars and colors *as they are now*.
        If later the widget gets updated via animation, updating label text,
        Layout's children being redrawn, etc., it will be un-flipped again.
        Second, most ASCII-art does not take it well. You're probably safe with
        something almost symmetrical, like background garbage and such, but for
        complex images it's better to provide both left and right versions.
        
        Unlike raster and vector graphics, there is no general way to flip an
        ASCII image programmatically. Except, of course, flipping tiles
        themselves which I find aesthetically unacceptable for my projects.
        
        :param axis:
        :return:
        """
        if axis in ('x', 'horizontal'):
            self.chars = [self.chars[x][::-1] for x in range(len(self.chars))]
            self.colors = [self.colors[x][::-1] for x in range(len(self.colors))]
        elif axis in ('y', 'vertical'):
            self.chars = self.chars[::-1]
            self.colors = self.colors[::-1]


class Layout(Widget):
    """
    A widget that can add others as its children.
    All children get drawn to its `chars` and are thus displayed on a single
    bearlibterminal layer. The layout does not explicitly pass events to its
    children, they are expected to subscribe to event queue by themselves.
    Children are allowed to overlap, but in that case the most recent one's char
    is actually drawn.
    The Layout is initialized with a single child, which is given chars and
    colors provided at Layout creation. This child is available as l.children[0]
    or as l.background
    The Layout automatically redraws itself on `tick` event, whether its
    children have updated or not.
    """
    def __init__(self, chars, colors):
        super().__init__(chars, colors)
        self.children = []
        # For every position, remember all the widgets that may want to place
        # characters in it, but draw only the latest one
        self._child_pointers = copy_shape(self.chars, None)
        # copy_shape does not work with lists correctly, so.
        for line in range(len(self._child_pointers)):
            for char in range(len(self._child_pointers[0])):
                self._child_pointers[line][char] = []
        self.child_locations = {}
        # The widget with Layout's chars and colors is created and added to the
        # Layout as the first child. It is done even if both are empty, just in
        # case someone wants to add background later
        w = Widget(self.chars, self.colors)
        self.add_child(w, pos=(0, 0))
    
    @property
    def terminal(self):
        return self._terminal
    
    # This setter propagates the terminal value to all the Layout's children.
    # It's necessary because some of them may be added before placing Layout on
    # the screen and thus end up terminal-less.
    @terminal.setter
    def terminal(self, value):
        if value and not isinstance(value, BearTerminal):
            raise BearException('Only BearTerminal can be added as terminal')
        self._terminal = value
        for child in self.children:
            child.terminal = value
        
    # Operations on children
    def add_child(self, child, pos, skip_checks = False):
        """
        Add a widget as a child at a given (relative) position.
        The child has to be a Widget or a Widget subclass that haven't yet been
        added to this Layout and whose dimensions are less than or equal to the
        Layout's
        :param child:
        :return:
        """
        if not isinstance(child, Widget):
            raise BearLayoutException('Cannot add non-Widget to a Layout')
        if child in self.children and not skip_checks:
            raise BearLayoutException('Cannot add the same widget to layout twice')
        if len(child.chars) > len(self._child_pointers) or \
                len(child.chars[0]) > len(self._child_pointers[0]):
            raise BearLayoutException('Cannot add child that is bigger than a Layout')
        if len(child.chars) + pos[1] > len(self._child_pointers) or \
                len(child.chars[0]) + pos[0] > len(self._child_pointers[0]):
            raise BearLayoutException('Child won\'t fit at this position')
        if child is self:
            raise BearLayoutException('Cannot add Layout as its own child')
        self.children.append(child)
        self.child_locations[child] = pos
        child.terminal = self.terminal
        child.parent = self
        for y in range(len(child.chars)):
            for x in range(len(child.chars[0])):
                self._child_pointers[pos[1] + y][pos[0] + x].append(child)

    def remove_child(self, child, remove_completely=True):
        """
        Remove a child from a Layout.
        :param child: the child to remove
        :param remove_completely: if False, the child is only removed from the
        screen, but remains in the children list. This is not intended to be
        used and is included only to prevent self.move_child from messing with
        child order.
        :return:
        """
        if child not in self.children:
            raise BearLayoutException('Layout can only remove its child')
        # process pointers
        for y in range(len(child.chars)):
            for x in range(len(child.chars[0])):
                self._child_pointers[self.child_locations[child][1] + y] \
                        [self.child_locations[child][0] + x].remove(child)
        if remove_completely:
            del(self.child_locations[child])
            self.children.remove(child)
            child.terminal = None
            child.parent = None
    
    def move_child(self, child, new_pos):
        self.remove_child(child, remove_completely=False)
        self.add_child(child, pos=new_pos, skip_checks=True)
    
    # BG's chars and colors are not meant to be set directly
    @property
    def background(self):
        return self.children[0]
    
    @background.setter
    def background(self, value):
        if not isinstance(value, Widget):
            raise BearLayoutException('Only Widget can be added as background')
        if not shapes_equal(self.chars, value.chars):
            # chars and colors are always the same size
            raise BearLayoutException('Wrong Layout background size')
        for row in range(len(self.chars)):
            for column in range(len(self.chars[0])):
                self._child_pointers[row][column][0] = value
                # self.colors[row][column][0] = value
        del self.child_locations[self.children[0]]
        self.child_locations[value] = (0, 0)
        self.children[0] = value
        
    def _rebuild_self(self):
        """
        Build fresh chars and colors for self
        :return:
        """
        chars = copy_shape(self.chars, ' ')
        colors = copy_shape(self.colors, None)
        for line in range(len(chars)):
            for char in range(len(chars[0])):
                for child in self._child_pointers[line][char][::-1]:
                    # Addressing the correct child position
                    c = child.chars[line-self.child_locations[child][1]] \
                        [char - self.child_locations[child][0]]
                    if c != ' ':
                        # Spacebars are used as empty space and are transparent
                        chars[line][char] = c
                        break
                colors[line][char] = \
                    child.colors[line - self.child_locations[child][1]] \
                    [char - self.child_locations[child][0]]
        self.chars = chars
        self.colors = colors
    
    def on_event(self, event):
        """
        The Layout redraws itself on every frame
        :return:
        """
        if event.event_type == 'service' and event.event_value == 'tick_over':
            self._rebuild_self()
            if isinstance(self.parent, BearTerminal):
                self.terminal.update_widget(self)
    
    #Service
    def get_absolute_pos(self, relative_pos):
        """
        Get an absolute position (in terminal coordinates) for any location
        within self.
        :param relative_pos:
        :return:
        """
        self_pos = self.terminal.widget_locations(self).pos
        return self_pos[0]+relative_pos[0], self_pos[1]+relative_pos[1]


class ScrollBar(Widget):
    """
    A scrollbar to be used with ScrollableLayout. On creation accepts the
    following parameters:
    `orientation` -- one of 'vertical' or 'horizontal'
    `length` -- int
    'colors' -- two BLT-compatible colors for background and the moving thingy
    
    Does not accept input
    """
    def __init__(self, orientation='vertical', length=10,
                 colors=('gray', 'white')):
        if orientation not in ('vertical', 'horizontal'):
            raise BearException(
                'Orientation must be either vertical or horizontal')
        if orientation == 'vertical':
            # TODO: custom chars in ScrollBar
            chars = [['#'] for _ in range(length)]
        else:
            chars = [['#' for _ in range(length)]]
        self.length = length
        self.orientation = orientation
        self.bg_color = colors[0]
        self.bar_color = colors[1]
        colors = copy_shape(chars, self.bg_color)
        super().__init__(chars, colors)
        
    def show_pos(self, position, percentage):
        """
        Move the scrollbar.
        :param position: Float. The position of the top (or left) side of the
        scrollbar, as part of its length
        :param percentage: Float. The lengths of the scrollbar, as part of the
        total bar length
        :return:
        """
        # Not really effective, but still quicker than Layout would be
        # Single-widget bar gets redrawn only when called, while a Layout
        # would've redrawn every tick
        start = round(self.length*position)
        width = round(self.length*percentage)
        self.colors = copy_shape(self.chars, self.bg_color)
        if self.orientation == 'vertical':
            for i in range(start, start+width):
                self.colors[i][0] = self.bar_color
        else:
            for i in range(start, start+width):
                self.colors[0][i] = self.bar_color


class ScrollableLayout(Layout):
    """
    A Layout that can show only a part of its surface.
    
    Like a Layout, accepts `chars` and `colors` on creation, which should be the
    size of the entire layout, not the visible area. The latter is initialized
    by `view_pos` and `view_size` arguments.
    
    Works by overloading _rebuild_self to only show a part of child_pointers
    """
    def __init__(self, chars, colors,
                 view_pos=(0, 0), view_size=(10, 10)):
        super().__init__(chars, colors)
        if not 0 <= view_pos[0] <= self.width - view_size[0] \
                or not 0 <= view_pos[1] <= self.height - view_size[1]:
            raise BearLayoutException('Initial viewpoint outside ' +
                                      'ScrollableLayout')
        if not 0 < view_size[0] <= len(chars[0]) \
                or not 0 < view_size[1] <= len(chars):
            raise BearLayoutException('Invalid view field size')
        self.view_pos = view_pos[:]
        self.view_size = view_size[:]
        self._rebuild_self()
    
    def _rebuild_self(self):
        """
        Same as `Layout()._rebuild_self`, but all child positions are also
        offset by `view_pos`. Obviously, only `view_size[1]` lines
        `view_size[0]` long are set as `chars` and `colors`.
        :return:
        """
        chars = [[' ' for x in range(self.view_size[0])] \
                 for y in range(self.view_size[1])]
        colors = copy_shape(chars, None)
        for line in range(self.view_size[1]):
            for char in range(self.view_size[0]):
                for child in self._child_pointers[self.view_pos[1]+line] \
                                     [self.view_pos[0] + char][::-1]:
                    # Addressing the correct child position
                    c = child.chars[self.view_pos[1] + line-self.child_locations[child][1]] \
                        [self.view_pos[0] + char-self.child_locations[child][0]]
                    if c != ' ':
                        # Spacebars are used as empty space and are transparent
                        chars[line][char] = c
                        break
                colors[line][char] = \
                    child.colors[self.view_pos[1] + line - self.child_locations[child][1]] \
                    [self.view_pos[0] + char - self.child_locations[child][0]]
        self.chars = chars
        self.colors = colors
    
    def resize_view(self, new_size):
        # TODO: support resizing view.
        # This will require updating the pointers in terminal or parent layout
        pass
    
    def scroll_to(self, pos):
        """
        Move field of view to `pos`.
        
        Raises `BearLayoutException` on incorrect position
        :param pos: tuple of ints
        :return:
        """
        if not (len(pos) == 2 and all((isinstance(x, int) for x in pos))):
            raise BearLayoutException('Field of view position should be 2 ints')
        if not 0 <= pos[0] <= len(self._child_pointers[0]) - self.view_size[0] \
                or not 0 <= pos[1] <= len(self._child_pointers)-self.view_size[1]:
            raise BearLayoutException('Scrolling to invalid position')
        self.view_pos = pos
    
    def scroll_by(self, shift):
        """
        Move field of view by `shift[0]` to the right and by `shift[1]` down.
        
        Raises `BearLayoutException` on incorrect position
        :param shift: tuple of ints
        :return:
        """
        pos = (self.view_pos[0] + shift[0], self.view_pos[1] + shift[1])
        self.scroll_to(pos)
    

class InputScrollable(Layout):
    """
    A ScrollableLayout wrapper that accepts input events and supports the usual
    scrollable view bells and whistles. Like ScrollableLayout, accepts chars and
    colors the size of the *entire* layout and inits visible area using view_pos
    and view_size.
    
    If bottom_bar and/or right_bar is set to True, it will be made one char
    bigger than view_size in the corresponding dimension to add ScrollBar.
    
    Can be scrolled by arrow keys.
    """
    def __init__(self, chars, colors, view_pos=(0, 0), view_size=(10, 10),
                 bottom_bar=False, right_bar=False):
        # Scrollable is initalized before self to avoid damaging it by the
        # modified view_size (in case of scrollbars)
        scrollable = ScrollableLayout(chars, colors, view_pos, view_size)
        size = list(view_size)
        ch = slice_nested(chars, view_pos, size)
        co = slice_nested(colors, view_pos, size)
        # Is there something more reasonable to add as ScrollableLayout BG?
        # It shouldn't be shown anyway
        if bottom_bar:
            ch.append(copy_shape(ch[0], ch[0][0]))
            co.append(copy_shape(co[0], co[0][0]))
        if right_bar:
            for x in ch:
                x.append(' ')
            for x in co:
                x.append('white')
        # While True, can add children to self. Otherwise they are passed to
        # self.scrollable
        self.building_self = True
        super().__init__(ch, co)
        self.scrollable = scrollable
        self.add_child(self.scrollable, pos=(0, 0))
        # Need to rebuild now to let bars know the correct height and width
        self._rebuild_self()
        if right_bar:
            self.right_bar = ScrollBar(orientation='vertical',
                                       length=self.height)
            self.add_child(self.right_bar, pos=(self.width-1, 0))
        else:
            self.right_bar = None
        if bottom_bar:
            self.bottom_bar = ScrollBar(orientation='horizontal',
                                        length=self.width)
            self.add_child(self.bottom_bar, pos=(0, self.height-1))
        self.building_self = False
        
    # TODO: accept mouse wheel and drag scroll
    def on_event(self, event):
        if event.event_type == 'key_down':
            scrolled = False
            if event.event_value == 'TK_DOWN' and \
              self.scrollable.view_pos[1] + self.scrollable.view_size[1]\
                    < len(self.scrollable._child_pointers):
                self.scrollable.scroll_by((0, 1))
                scrolled = True
            elif event.event_value == 'TK_UP' and \
             self.scrollable.view_pos[1] > 0:
                self.scrollable.scroll_by((0, -1))
                scrolled = True
            elif event.event_value == 'TK_RIGHT' and \
              self.scrollable.view_pos[0] + self.scrollable.view_size[0]\
                    < len(self.scrollable._child_pointers[0]):
                self.scrollable.scroll_by((1, 0))
                scrolled = True
            elif event.event_value == 'TK_LEFT' and \
              self.scrollable.view_pos[0] > 0:
                self.scrollable.scroll_by((-1, 0))
                scrolled = True
            elif event.event_type == 'TK_SPACE':
                self.scrollable.scroll_to((0, 0))
                scrolled = True
            if scrolled:
                if self.right_bar:
                    self.right_bar.show_pos(
                        self.scrollable.view_pos[1] /
                            len(self.scrollable._child_pointers),
                        self.scrollable.view_size[0] /
                            len(self.scrollable._child_pointers))
        super().on_event(event)

    def add_child(self, child, pos, skip_checks=False):
        if not self.building_self:
            self.scrollable.add_child(child, pos, skip_checks)
        else:
            super().add_child(child, pos, skip_checks)
            

# Animations and other complex decorative Widgets
class Animation:
    """
    A data class for animation, ie the sequence of the frames
    """
    def __init__(self, frames, fps):
        if not all((shapes_equal(x[0], frames[0][0]) for x in frames[1:])) \
                or not all(
                (shapes_equal(x[1], frames[0][1]) for x in frames[1:])):
            raise BearException('Frames should be equal size')
        self.frames = frames
        self.frame_time = 1 / fps

    def __len__(self):
        return len(self.frames)


class SimpleAnimationWidget(Widget):
    """
    A simple animated widget that cycles through the frames.

    Accepts two parameters on creation:
    `frames` an iterable of (chars, colors) tuples. These should all be
    the same size
    `fps` frames per second.
    `emit_ecs`: whether to emit ecs_update events on every frame. Useless for
    widgets outside ecs system, but those on ECSLayout are not redrawn unless
    this event is emitted or something else causes ECSLayout to redraw
    """
    
    def __init__(self, animation, emit_ecs=True):
        if not isinstance(animation, Animation):
            raise BearException(
                'Only Animation instance can be used in SimpleAnimationWidget')
        self.animation = animation
        super().__init__(*animation.frames[0])
        self.running_index = 0
        self.have_waited = 0
        self.emit_ecs = emit_ecs
    
    def on_event(self, event):
        if event.event_type == 'tick':
            self.have_waited += event.event_value
            if self.have_waited >= self.animation.frame_time:
                self.running_index += 1
                if self.running_index >= len(self.animation):
                    self.running_index = 0
                self.chars = self.animation.frames[self.running_index][0]
                self.colors = self.animation.frames[self.running_index][1]
                self.have_waited = 0
                if self.emit_ecs:
                    return BearEvent(event_type='ecs_update')
        elif self.parent is self.terminal and event.event_type == 'service' \
                and event.event_value == 'tick_over':
            # This widget is connected to the terminal directly and must update
            # itself without a layout
            self.terminal.update_widget(self)


class MultipleAnimationWidget(Widget):
    """
    A widget that is able to display multiple animations.
    """
    def __init__(self, animations, initial_animation,
                 emit_ecs=True, cycle=False):
        # Check the animations' validity
        if not isinstance(animations, dict) or \
                any((not isinstance(x, Animation) for x in animations.values())):
            raise BearException(
                'Only dict of Animations acceptable for MultipleAnimationWidget')
        if any((not isinstance(x, str) for x in animations)):
            raise BearException('Animation names should be strings')
        if not initial_animation:
            raise BearException('Initial animation ID should be provided')
        if initial_animation not in animations:
            raise BearException('Incorrect initial animation ID')
        super().__init__(*animations[initial_animation].frames[0])
        self.animations = animations
        self.current_animation = initial_animation
        self.running_index = 0
        self.have_waited = 0
        self.emit_ecs = emit_ecs
        self.cycle = cycle
        self.am_running = True

    def on_event(self, event):
        # When self.am_running is False, this widget does not respond to any
        # events and acts like a regular passive Widget
        if self.am_running:
            if event.event_type == 'tick':
                self.have_waited += event.event_value
                if self.have_waited >= self.animation.frame_time:
                    self.running_index += 1
                    if self.running_index >= len(self.animation):
                        if self.cycle:
                            self.running_index = 0
                        else:
                            self.am_running = False
                    self.chars = self.animation.frames[self.running_index][0]
                    self.colors = self.animation.frames[self.running_index][1]
                    self.have_waited = 0
                    if self.emit_ecs:
                        return BearEvent(event_type='ecs_update')
            elif self.parent is self.terminal and event.event_type == 'service'\
                    and event.event_value == 'tick_over':
                # This widget is connected to the terminal directly and must
                # update itself without a layout
                self.terminal.update_widget(self)

    @property
    def animation(self):
        return self.animations[self.current_animation]

    def set_animation(self, anim_id, cycle=False):
        """
        Set the next animation to be played.
        :param anim_id: Animation ID. Should be present in self.animations
        :param cycle: Whether to cycle the animation. Default False.
        :return:
        """
        if anim_id not in self.animations:
            raise BearException('Incorrect animation ID')
        self.current_animation = anim_id
        self.cycle = cycle
        self.am_running = True


# Functional widgets. Please note that these include no decoration, BG, frame or
# anything else. Ie Label is just a chunk of text on the screen, FPSCounter and
# MousePosWidget are just the numbers that change. For the more complex visuals,
# embed these into a Layout with a preferred BG

class Label(Widget):
    """
    A widget that displays text.
    Accepts only a single string, whether single- or multiline.
    Does not (yet) support complex text markup used by bearlibterminal
    :param text: string to be displayed
    :param just: horizontal text justification, one of 'left', 'right'
    or 'center'. Default 'left'.
    :param color: bearlibterminal-compatible color. Default 'white'
    :param width: text area width. Defaults to the length of the longest
    substring in `text`.
    :param height: text area height. Defaults to the line count in `text`

    Label's text can be edited at any time by setting label.text property. Note
    that it overwrites any changes to `self.chars` and `self.colors` made after
    setting `self.text` the last time.
    
    Unlike text, Label's height and width cannot be changed. Set its height and
    width to accomodate all possible inputs during Label creation.
    """
    
    def __init__(self, text,
                 just='left', color='white', width=None, height=None):
        chars = Label._generate_chars(text, width, height, just)
        colors = copy_shape(chars, color)
        super().__init__(chars, colors)
        self.color = color
        # Bypassing setter, because I need to actually create fields
        self._just = just
        self._text = text
    
    @staticmethod
    def _generate_chars(text, width, height, just):
        """
        Internal method that generates a justified char list for the Label
        :param text:
        :param just:
        :return:
        """
        
        def justify(line, width, just_type):
            if len(line) < width:
                if just_type == 'left':
                    return line.ljust(width)
                elif just_type == 'right':
                    return line.rjust(width)
                elif just_type == 'center':
                    lj = width - int((width - len(line)) / 2)
                    return line.ljust(lj).rjust(width)
                else:
                    raise BearException(
                        'Justification should be \'left\', \'right\' or \'center\'')
            else:
                return line
        
        lines = text.split('\n')
        if not width:
            width = max(len(x) for x in lines)
        r = [list(justify(x, width, just)) for x in lines]
        if height and len(r) < height:
            for x in range(height - len(r)):
                r.append([' ' for j in range(len(r[0]))])
        return r

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        if not self._text:
            self._text = value
        chars = copy_shape(self.chars, ' ')
        self.chars = blit(chars, self._generate_chars(value,
                                                      len(self.chars[0]),
                                                      len(self.chars),
                                                      self.just),
                          0, 0)
        self._text = value
        # MousePosWidgets (a child of Label) may have self.terminal set
        # despite not being connected to the terminal directly
        if self.terminal and self in self.terminal._widget_pointers:
            self.terminal.update_widget(self)

    @property
    def just(self):
        return self._just
    
    @just.setter
    def just(self, value):
        self.chars = Label._generate_chars(self.text, len(self.chars[0]),
                                           len(self.chars), just=value)
        if self.terminal:
            self.terminal.update_widget(self)
            
            
class InputField(Label):
    """
    A single-line field for keyboard input.
    
    The length of the input line is limited by the InputField size. When the
    input is finished (by pressing ENTER), InputField emits a
    `BearEvent(event_type='text_input', event_value=(field.name, field.text))`
    Since BLT has no support for system keyboard layouts, only supports Latin.
    This also applies to non-letter symbols: for example, comma and period are
    considered to be on different keys even in Russian layout, where they are
    on the same one.
    """
    charcodes = {'SPACE': ' ', 'MINUS': '-', 'EQUALS': '=',
                 'LBRACKET': '[', 'RBRACKET': ']', 'BACKSLASH': '\\',
                 'SEMICOLON': ';', 'APOSTROPHE': '\'', 'GRAVE': '`',
                 'COMMA': ',', 'PERIOD': '.', 'SLASH': '/',
                 'KP_DIVIDE': '/', 'KP_MULTIPLY': '*', 'KP_MINUS': '-',
                 'KP_PLUS': '+', 'KP_1': '1', 'KP_2': '2', 'KP_3': '3',
                 'KP_4': '4', 'KP_5': '5', 'KP_6': 6, 'KP_7': 7,
                 'KP_8': 8, 'KP_9': 9, 'KP_0': 0, 'KP_PERIOD': '.'
                 }
    
    # Charcodes for non-letter characters used via Shift button
    shift_charcodes = {'MINUS': '_', 'EQUALS': '+', 'LBRACKET': '{',
                       'RBRACKET': '}', 'BACKSLASH': '|', 'SEMICOLON': ':',
                       'APOSTROPHE': '\"', 'GRAVE': '~', 'COMMA': '<',
                       'PERIOD': '>', 'SLASH': '?', '1': '!', '2': '@',
                       '3': '#', '4': '$', '5': '%', '6': '^', '7': '&',
                       '8': '*', '9': '(', '0': ')'}
    
    def __init__(self, name='Input field', accept_input=True, **kwargs):
        if 'width' not in kwargs:
            raise BearException('InputField cannot be created without ' +
                                'either `width` or default text')
        super().__init__('', **kwargs)
        # The name will be used when the input is finished
        self.name = name
        self.shift_pressed = False
        # Set to True to return 'text_input' and
        self.finishing = False
        self.accept_input = accept_input
        
    def on_event(self, event):
        #TODO Reactivate InputField on mouse click, if inactive
        # Requires it to have terminal for state.
        if self.finishing:
            # If finishing, the event will be ignored
            return BearEvent(event_type='text_input',
                             event_value=(self.name, self.text))
        if self.accept_input and event.event_type == 'key_down':
            # Stripping 'TK_' part
            symbol = event.event_value[3:]
            if symbol == 'BACKSPACE':
                self.text = self.text[:-1]
            elif symbol == 'SHIFT':
                self.shift_pressed = True
            # TK_ENTER is presumed to be the end of input
            elif symbol == 'ENTER':
                self.accept_input = False
                # returns immediately, unlike self.finish which sets it to
                # return on the *next* event
                return BearEvent(event_type='text_input',
                                 event_value=(self.name, self.text))
            elif len(self.text) < len(self.chars[0]):
                self.text += self._get_char(symbol)
            if self.terminal:
                self.terminal.update_widget(self)
        elif event.event_type == 'key_up':
            if event.event_value == 'TK_SHIFT':
                self.shift_pressed = False

    def finish(self):
        """
        Finish accepting the input and emit the 'text_input' event at the next
        opportunity.
        
        The event will not be actually emitted until the next event is passed
        to on_event.
        :return:
        """
        self.accept_input = False
        self.finishing = True
        
    def _get_char(self, symbol):
        """
        Return the char corresponding to a TK_* code.
        
        Considers the shift state
        :param symbol:
        :return:
        """
        if len(symbol) == 1:
            if self.shift_pressed:
                if symbol in '1234567890':
                    return self.shift_charcodes[symbol]
                else:
                    return symbol
            else:
                return symbol.lower()
        elif symbol in self.charcodes:
            if self.shift_pressed and symbol in self.shift_charcodes:
                return self.shift_charcodes[symbol]
            else:
                return self.charcodes[symbol]
        else:
            return ''


class FPSCounter(Label):
    """
    A simple widget that measures FPS.
    Actually just prints 1/(average runtime over the last 100 ticks in seconds),
    so it takes 100 ticks to get an accurate reading. Not relevant except on the
    first several seconds of the program run or after FPS has changed, but if it
    seems like the game takes a second or two to reach the target FPS -- it just
    seems that way.
    """
    def __init__(self, **kwargs):
        self.samples_deque = deque(maxlen=100)
        super().__init__('030', **kwargs)
    
    def _update_self(self):
        fps = str(round(len(self.samples_deque) /
                            sum(self.samples_deque)))
        fps = fps.rjust(3, '0')
        self.text = fps
    
    def on_event(self, event):
        # Update FPS estimate
        if event.event_type == 'tick':
            self.samples_deque.append(event.event_value)
            self._update_self()
            if self.parent is self.terminal:
                self.terminal.update_widget(self, refresh=True)


class MousePosWidget(Label):
    """
    A simple widget akin to FPSCounter that listens to TK_MOUSE_MOVE events.
    
    In order to work, it needs `self.terminal` to be set to the current
    terminal, which means it should either be added to the terminal directly
    (without any Layouts) or it should be set manually before MousePosWidget
    gets its first `tick` event.
    """
    
    def __init__(self, **kwargs):
        super().__init__(text='000x000', **kwargs)
        
    def on_event(self, event):
        if event.event_type == 'misc_input' and \
                     event.event_value == 'TK_MOUSE_MOVE':
            self.text = self.get_mouse_line()
        if isinstance(self.parent, BearTerminal):
            self.terminal.update_widget(self)

    def get_mouse_line(self):
        if not self.terminal:
            raise BearException('MousePosWidget is not connected to a terminal')
        x = str(self.terminal.check_state('TK_MOUSE_X')).rjust(3, '0')
        y = str(self.terminal.check_state('TK_MOUSE_Y')).rjust(3, '0')
        return x + 'x' + y
  
    
# Listeners
class Listener:
    """
    A base class for the things that need to interact with the queue (and maybe
    the terminal), but aren't widgets.
    """
    def __init__(self):
        self.terminal = None
    
    def on_event(self, event):
        raise NotImplementedError('Listener base class is doing nothing')
    
    def register_terminal(self, terminal):
        """
        Register a terminal with which this listener will interact
        :param terminal: A BearTerminal instance
        :return:
        """
        if not isinstance(terminal, BearTerminal):
            raise TypeError('Only BearTerminal instances registered by Listener')
        self.terminal = terminal
        
    
class ClosingListener(Listener):
    """
    The listener that waits for TK_CLOSE input event (Alt-F4 or closing window)
    and sends the shutdown service event to the queue. All widgets are expected
    to listen to it and immediately save their data or whatever they need to do.
    On the next tick ClosingListener closes the terminal and queue altogether.
    """
    def __init__(self):
        super().__init__()
        self.countdown = 2
        self.counting = False
        
    def on_event(self, event):
        if event.event_type == 'misc_input' and event.event_value == 'TK_CLOSE':
            self.counting = True
            return BearEvent(event_type='service', event_value='shutdown_ready')
        if event.event_type == 'tick':
            if self.counting:
                self.countdown -= 1
                if self.countdown == 0:
                    return BearEvent(event_type='service',
                                     event_value='shutdown')


class LoggingListener(Listener):
    """
    A listener that logs the events it accepts
    """
    def __init__(self, handle):
        super().__init__()
        if not hasattr(handle, 'write'):
            raise BearException('The LoggingListener needs a writable object')
        self.handle = handle
        
    def on_event(self, event):
        self.handle.write('{0}: type {1}, '.format(str(time()), event.event_type) +
                          'value {}\n'.format(event.event_value))



from setuptools import setup
setup(
  name = 'bear_hug',
  packages = ['bear_hug'],
  version = '0.1.1',
  description = 'An object-oriented ECS library for ASCII games. Wraps bearlibterminal.',
  author = 'Alexey Morozov',
  author_email = 'alexeymorozov1991@gmail.com',
  url = 'https://github.com/SynedraAcus/bear_hug',
  download_url = 'https://github.com/SynedraAcus/bear_hug/archive/0.1.1.tar.gz',
  keywords = ['gamedev', 'ascii', 'ASCII'],
  license = 'MIT',
  python_requires = '>=3.6',
  install_requires=['bearlibterminal'],
  classifiers = [],
)



"""
An event system.
All events are passed around on every tick to listeners' `on_event` method according
to their `event_type` subscriptions.
`on_event` may return either a BearEvent or a list of BearEvents to be added to
the queue. If a list is returned, the events are added (and will be processed)
in the same order as they are in that list.
"""

from bear_hug.bear_utilities import BearLoopException
from collections import deque


class BearEvent:
    """
    Event data class. Has two params: event_type and event_value
    """
    def __init__(self, event_type='tick', event_value=None):
        self.event_type = event_type
        self.event_value = event_value
        

class BearEventDispatcher:
    """
    The BearEvent queue and dispatcher class.
    Iterates until someone emits the 'shutdown' event of type 'service'. Widgets
    may expect the 'shutdown_ready' event of the same type to be emitted a tick
    before that, so that they could finish processing the last tick and save
    their data or whatever. But this is not enforced by the queue.
    """
    def __init__(self):
        self.last_tick_time = None
        # All these event types need to be supported by the queue in order for
        # the engine to function.
        self.event_types = {'tick', # Emitted every tick
                            'key_down', # Key or mouse button down
                            'key_up', # Key or mouse button up
                            'misc_input', # Other input, eg MOUSE_MOVE or CLOSE
                            'text_input', #InputField returns something
                            'play_sound', # Something requests a sound
                            'service', # To do with queue or engine in general
                            'ecs_create', # A creation of entity
                            'ecs_move', # Movement of entities
                            'ecs_collision', # Collision of entities
                            'ecs_add', # Addition of entity to ECSLayout
                            'ecs_remove', # Removal of entites from ECSLayout
                            'ecs_update' # Someone needs to update ecs screen
                            }
        self.listeners = {x: [] for x in self.event_types}
        self.deque = deque()
    
    def register_listener(self, listener, event_types='all'):
        """
        Add a listener to this event_dispatcher.
        :param object listener: a listener to add. Any object with an `on_event`
        callback can be added. The callback should accept a BearEvent instance
        as a single parametickter.
        :param iterable|str event_types: either a string or an iterable of
        strings. If an iterable, its elements are event types the listener
        subscribes to.
        
        If a string, the following rules apply:
        * If a string equals 'all', the listener is subscribed to all currently
        registered event types.
        * Elif a string starts with '*', the listener is subscribed to all
        currently registered event types for whose name event_types[1:] is a
        substring.
        * Else a string is interpreted as a single event type.
        
        In all cases incorrect event types raise BearLoopException.
        Defaults to 'all'
        :return:
        """
        if not hasattr(listener, 'on_event'):
            raise BearLoopException('Cannot add an object without on_event' +
                                    ' method as a listener')
        if isinstance(event_types, str):
            if event_types == 'all':
                # Subscribing to all events
                types = self.listeners.keys()
            elif event_types[0] == '*':
                # Subscribing to a group of events
                types = []
                mask = event_types[1:]
                for event_type in self.listeners:
                    if mask in event_type:
                        types.append(event_type)
            else:
                # Subscribing to a single event type
                types = [event_types]
        else:
            # Subscribing to a list of event types
            types = event_types
        for event_type in types:
            try:
                self.listeners[event_type].append(listener)
            except KeyError:
                # The incorrect list elements or single value processed here
                raise BearLoopException('Unknown event class {}'.format(
                                            event_type))
    
    def unregister_listener(self, listener, event_types='all'):
        """
        Remove a listener from the event_dispatcher or some of its event types.
        :param object listener: listener to remove
        :param iterable|str event_types: event types to unsubscribe from or
        'all'. Defaults to 'all'
        :return:
        """
        if event_types == 'all':
            event_types = self.listeners.keys()
        for event_type in event_types:
            if listener in self.listeners[event_type]:
                self.listeners[event_type].remove(listener)
                
    def register_event_type(self, event_type):
        """
        Add a new event type to be processed by queue.
        
        This makes passing (and subscribing to) a new event type possible. No
        listeners are automatically subscribed to it, even those that were
        initially registered with 'all' or fitting '*'-types.
        :param event_type:
        :return:
        """
        if not isinstance(event_type, str):
            raise ValueError('Event type must be a string')
        self.event_types.add(event_type)
        self.listeners[event_type] = []
    
    def add_event(self, event):
        """
        Add a BearEvent to the queue.
        :param event:
        :return:
        """
        if not isinstance(event, BearEvent):
            raise BearLoopException('Only BearEvents can be added to queue')
        if event.event_type not in self.event_types:
            raise BearLoopException('Incorrect event type \"{}\"'.format(
                event.event_type))
        self.deque.append(event)
    
    def start_queue(self):
        """
        Sends the queue initialization event to deque
        :return:
        """
        self.deque.append(BearEvent(event_type='service',
                                    event_value='Queue started'))
    
    def dispatch_events(self):
        """
        Dispatch all the events to their listeners, adding whatever they have to
        say about it to the queue.
        :return:
        """""
        while len(self.deque) > 0:
            e = self.deque.popleft()
            for listener in self.listeners[e.event_type]:
                r = listener.on_event(e)
                if r:
                    if isinstance(r, BearEvent):
                        self.add_event(r)
                    elif isinstance(r, list):
                        for event in r:
                            self.add_event(event)
                    else:
                        raise BearLoopException('on_event returns something ' +
                                                'other than BearEvent')
